#pragma warning(disable:4244)
#pragma warning(disable:4018)

#include "definitions.h"
#include "textoutput.h"
#include "precisetimer.h"
#include "particle.h"
#include "tetramesh.h"

DclVector getWorld();
void  loadTetraFile(Mesh& m, char *filename);
void  computeSpringSet(Mesh& m);
void  updateParticles(Mesh& m);


const int MAX_NUM_PARTICLES = 10000;

// Integration step h
const DclFloat  TIMESTEP = 0.005;

// Elasticity of collision with walls
const DclFloat  WORLD_ELASTICITY = 1.0;

const DclFloat  SPRING_CONSTANT = 20000.0;
const DclFloat  SPRING_DAMPING_CONSTANT = 10.0;
const DclFloat  FRICTION = 100;

const Integration INTEGRATION = RungeKutta4;

//////////// GLOBALS /////////////////
std::vector<Mesh> vectorMeshes;

int frame = 0;

//////////////////////////////////////

void  visualizeParticles();

//------------------------------------------------------------------------------------------------------------------------------------

void  initializeTetraMesh(Mesh& m, DclVector translation)
{
  // initialize particles
  for (int p = 0; p < m.vectorParticles.size(); p ++)
  {
    Particle& particle = m.vectorParticles[p];
    particle.position += translation;
    particle.positionInit = particle.position;
    particle.mass = 1.0;
    particle.r = m.r; particle.g = m.g; particle.b = m.b;
  }
}

void  initializeRigidBody(Mesh& m)
{
  m.position.clear();
  m.mass = 0;

  // sum the masses of the particles to obtain the total rigid body mass and the center of mass
  for (int p = 0; p < m.vectorParticles.size(); p ++)
  {
    Particle& particle = m.vectorParticles[p];
    m.position += particle.position * particle.mass;
    m.mass += particle.mass;
  }

  m.position /= m.mass;
  m.positionInit = m.position;

  // compute the inertia tensor
  m.I.clear();
  for (int p = 0; p < m.vectorParticles.size(); p ++)
  {  
    Particle& particle = m.vectorParticles[p];
    DclVector r = particle.position - m.position;
    // skew matrix of r
    DclMatrix33 skew_r(DclVector(0, r.z, -r.y), DclVector(-r.z, 0, r.x), DclVector(r.y, -r.x, 0));
    m.I = m.I - (skew_r * skew_r) * particle.mass;
  }
  // inverted inertia tensor
  m.Iinv = m.I;
  m.Iinv.inverse();


  // obtain an arbitrary orthonormal orientation matrix
  DclVector right(1, 0, 1);
  right.normalize();
  DclVector back(0, 1, 1);
  DclVector up = right % back;
  back = up % right;
  right.normalize();
  back.normalize();
  up.normalize();

  m.orientation = DclMatrix33(right, back, up);
}


//------------------------------------------------------------------------------------------------------------------------------------

// initialize the particle properties
void  initializeSimulation(int argc, char ** argv)
{
  std::string filename;
  if (argc > 1) 
  {
    filename = argv[1];
  }
  else
  {
    filename = "cube.dcstructure";
  }
  vectorMeshes.push_back(Mesh());

  Mesh& m1 = vectorMeshes[0];
  m1.integration = Euler;
  loadTetraFile(m1, (char*)filename.c_str());
  computeSpringSet(m1);
  initializeTetraMesh(m1, DclVector(0, 0, 0));

  initializeRigidBody(m1);
  updateParticles(m1);
}

//------------------------------------------------------------------------------------------------------------------------------------

// add external forces to the sum of forces acting on the particle
void  addForces(Mesh& m)
{
  m.force += DclVector(0, 1.0, 0) * m.mass * -9.81; // gravitational force
}

//------------------------------------------------------------------------------------------------------------------------------------

// perform one time-integration step
void  integrate(Mesh& m)
{
  // sum the forces and torques of all particles to obtain the total force and torque
  for (int i = 0; i < m.vectorParticles.size(); i ++)
  {
    Particle& particle = m.vectorParticles[i];
    m.force += particle.force;
    m.torque += (particle.position - m.position) % particle.force;
  }

  DclVector positionOld = m.position;
  // Euler-Cromer scheme
  m.velocity += m.force / m.mass * TIMESTEP;
  m.position += m.velocity * TIMESTEP;

  // compute the transposed orientation matrix
  DclMatrix33 AT(m.orientation);
  AT.transpose();

  // compute the term A * I-1 * AT tau
  DclVector ATtau = AT * m.torque;
  DclVector JATtau = m.Iinv * ATtau;
  DclVector AJATtau = m.orientation * JATtau;

  // numerically integrate the angular velocity
  m.angularVelocity += AJATtau * TIMESTEP;

  // skew matrix of the angular velocity
  DclMatrix33 skew_omega(
    DclVector(0, m.angularVelocity.z, -m.angularVelocity.y), 
    DclVector(-m.angularVelocity.z, 0, m.angularVelocity.x), 
    DclVector(m.angularVelocity.y, -m.angularVelocity.x, 0));

  // numerically integrate the orientation
  m.orientation = m.orientation + skew_omega * m.orientation * TIMESTEP;

  // re-orthonormalize the orientation matrix
  DclVector a1 = DclVector(m.orientation.m[0][0], m.orientation.m[1][0], m.orientation.m[2][0]);
  DclVector a2 = DclVector(m.orientation.m[0][1], m.orientation.m[1][1], m.orientation.m[2][1]);
  DclVector a3 = DclVector(m.orientation.m[0][2], m.orientation.m[1][2], m.orientation.m[2][2]);
  DclVector b1 = a1; b1.normalize();
  DclVector b2 = a2 - b1 * (b1 * a2); b2.normalize();
  DclVector b3 = b1 % b2; b3.normalize();
  m.orientation = DclMatrix33(b1, b2, b3);
}

//------------------------------------------------------------------------------------------------------------------------------------

// collision handling with the world
void  worldCollision(Mesh &m, Particle& particle)
{
  DclVector world = getWorld();

  world *= 0.5;

  DclVector r = particle.position - m.position;
  DclVector penetrationDepth(0, 0, 0);

  // collision test with left and right wall
  if (particle.position.x < -world.x || particle.position.x > world.x)
  {
    penetrationDepth = DclVector(SIGN(particle.position.x) * world.x - particle.position.x, 0, 0);
  }

  // collision test with floor and ceiling
  if (particle.position.y < -world.y || particle.position.y > world.y)
  {
    penetrationDepth = DclVector(0, SIGN(particle.position.y) * world.y - particle.position.y, 0);
  }

  // collision test with front and back wall
  if (particle.position.z < -world.z || particle.position.z > world.z)
  {
    penetrationDepth = DclVector(0, 0, SIGN(particle.position.z) * world.z - particle.position.z);
  }

  // this mass-point is collided
  if (penetrationDepth.length2())
  {
    // the contact space (or contact normal) is d / |d|
    DclVector contactSpace(penetrationDepth); contactSpace.normalize();

    // the penalty force of this mass point
    particle.force += penetrationDepth * SPRING_CONSTANT;

    // velocity of point
    DclVector  v_point = m.velocity + m.angularVelocity % r;

    // velocity of point perpendicular to collision plane
    DclVector v_perp = v_point.projectOn(contactSpace);

    // damping force
    particle.force -= v_perp * SPRING_DAMPING_CONSTANT;

    // velocity of point parallel to collision plane
    DclVector v_tangent = v_point - v_perp;

    // friction force
    particle.force -= v_tangent * FRICTION;
  }
}

//------------------------------------------------------------------------------------------------------------------------------------

// compute and output the total energies in the scene
void  computeEnergies(Mesh& m, int i)
{
  DclFloat  e_kin = 0, e_pot = 0;
  DclVector world = getWorld();

  // sum the potential energy of each point
  for (int part = 0; part < m.vectorParticles.size(); part ++)
  {
    Particle& particle = m.vectorParticles[part];

    e_pot += (particle.position.y + world.z * 0.5) * particle.mass * 9.81;
  }

  // kinetic energy of cm
  e_kin = m.velocity.length2() * m.mass * 0.5;
  // rotational energy
  DclFloat  e_rot = m.angularVelocity * (m.I * m.angularVelocity) * 0.5;

  Textout::printvar(6 + i, 0, "Ekin %f, Epot %f, Erot %f, Sum %f", e_kin, e_pot, e_rot, e_kin + e_pot + e_rot);
}


void  updateParticles(Mesh& m)
{
  // update the positions of each particle based on the orientation of the rigid body
  for (int part = 0; part < m.vectorParticles.size(); part ++)
  {
    DclVector d = m.vectorParticles[part].positionInit - m.positionInit;
    d = m.orientation * d;
    m.vectorParticles[part].position = d;
    m.vectorParticles[part].position += m.position;
  }

}

//------------------------------------------------------------------------------------------------------------------------------------

// perform one simulation step
void  simulationStep()
{
  static int splitCounter = 0;

  int numParticles = 0;

  for (int mesh = 0; mesh < vectorMeshes.size(); mesh ++)
  {
    Mesh& m = vectorMeshes[mesh];

    // clear the forces of the particles
    for (int part = 0; part < m.vectorParticles.size(); part ++)
    {
      m.vectorParticles[part].force.clear();
    }
    m.force.clear();
    m.torque.clear();

    // compute external forces on the particles
    addForces(m);

    // compute collision impulses
    for (int part = 0; part < m.vectorParticles.size(); part ++)
    {
      worldCollision(m, m.vectorParticles[part]);
    }

    integrate(m);

    // compute and output energies
    computeEnergies(m, mesh);
    // update the position of the particles
    updateParticles(m);

  }

  // draw particles
  visualizeParticles();

  // measure time for frame
  double  dt = PreciseTimer::stop("SimulationStep");
  PreciseTimer::start("SimulationStep");

  // output statistics
  Textout::printvar(10, 0, "Frame %d, Frames/second %f, Num. particles %d", frame, 1000.0 / dt, numParticles);
  frame ++;
}


