/******************************************************************************
 * quaternion.h
 * Copyright (C) 2004 Bruno Heidelberger <heidelberger@inf.ethz.ch>
 *                    Matthias Teschner <teschner@inf.ethz.ch>
 ******************************************************************************/

#ifndef DCL_QUATERNION_H
#define DCL_QUATERNION_H

//****************************************************************************//
// Forward declaration
//****************************************************************************//

class DclVector;

//****************************************************************************//
// Class declaration
//****************************************************************************//

 /*****************************************************************************/
/** The quaternion class.
  *****************************************************************************/

class DclQuaternion
{
// member variables
public:
  union {
    DclFloat x;
    DclFloat q1;
  };
  union {
    DclFloat y;
    DclFloat q2;
  };
  union {
    DclFloat z;
    DclFloat q3;
  };
  union {
    DclFloat w;
    DclFloat q4;
  };

// constructors/destructor
public:
  DclQuaternion();
  DclQuaternion(const DclQuaternion& q);
  DclQuaternion(const DclVector& v, DclFloat a);
  DclQuaternion(DclFloat qx, DclFloat qy, DclFloat qz, DclFloat qw);
  /// Constructs a quaternion from a right-handed orthonormal coordinate system given by e1, e2 and e3. The user is responsible for normalizing e1, e2 and e3.
  DclQuaternion(DclVector& e1, DclVector& e2, DclVector& e3);
  ~DclQuaternion();

// member functions	
public:
  void blend(DclFloat d, const DclQuaternion& q);
  void clear();
  void conjugate();
  void invert();
  void normalize();
  void nullize();
  void set(const DclVector& v, DclFloat a);
  void set(DclFloat qx, DclFloat qy, DclFloat qz, DclFloat qw);


  /*****************************************************************************/
  /** Multiplies another quaternion to the quaternion instance.
  *
  * This operator multiplies another quaternion to the quaternion instance.
  *
  * @param q The quaternion to be multiplied.
  *****************************************************************************/
  void operator*=(const DclQuaternion& q);


  /*****************************************************************************/
  /** Multiplies a vector to the quaternion instance.
  *
  * This operator multiplies a vector to the quaternion instance.
  *
  * @param v The vector to be multiplied.
  *****************************************************************************/
  void operator*=(const DclVector& v);

  /*****************************************************************************/
  /** Multiplies a scalar to the quaternion instance.
  *
  * @param f The scalar to be multiplied.
  *****************************************************************************/
  void operator*=(const DclFloat& f);

  /*****************************************************************************/
  /** Adds two quaternions together.
  *
  * This method is used for numerical integration, q = q + dq * dt
  *
  * @param dq The quaternion to be added
  *****************************************************************************/
  void operator+=(const DclQuaternion& dq);

  DclFloat& operator[](unsigned int i);
  const DclFloat& operator[](unsigned int i) const;
  void operator=(const DclQuaternion& q);
  friend DclQuaternion operator*(const DclQuaternion& q, const DclQuaternion& r);
  friend DclQuaternion operator*(const DclQuaternion& q, const DclFloat s);
  friend DclQuaternion operator+(const DclQuaternion& q, const DclQuaternion& r);

  /*****************************************************************************/
  /** Multiplies a quaternion to the quaternion instance, r = Q * q;
  * 
  *****************************************************************************/
  DclQuaternion multiply(const DclQuaternion& q);
  inline void multiply(DclQuaternion& q_target, DclFloat q1, DclFloat q2, DclFloat q3, DclFloat q4);

  /*****************************************************************************/
  /** Multiplies a quaternion to the transposed quaternion instance, r = Q^T * q;
  *
  *****************************************************************************/
  DclQuaternion multiplyTranspose(const DclQuaternion& q);
  inline void multiplyTranspose(DclQuaternion& q_target, DclFloat q1, DclFloat q2, DclFloat q3, DclFloat q4);

  /*****************************************************************************/
  /** Multiplies a quaternion to the inverted quaternion instance, r = Q^-1 * q;
  *
  *****************************************************************************/
  DclQuaternion multiplyInverse(const DclQuaternion& q);
  inline void multiplyInverse(DclQuaternion& q_target, DclFloat q1, DclFloat q2, DclFloat q3, DclFloat q4);

  /*****************************************************************************/
  /** Multiplies a quaternion to the inverted transposed quaternion instance, r = Q^-1^T * q;
  *
  *****************************************************************************/
  DclQuaternion multiplyInverseTranspose(const DclQuaternion& q);
  inline void multiplyInverseTranspose(DclQuaternion& q_target, DclFloat q1, DclFloat q2, DclFloat q3, DclFloat q4);



  /*****************************************************************************/
  /** Returns a orthogonal right-handed vector triplet representing the direction of the quaternion, scaled by scale.
  *
  *****************************************************************************/
  void toDirector(DclVector& d1, DclVector& d2, DclVector& d3, DclFloat scale = 1.0);

  /*****************************************************************************/
  /** Returns the ith director representing the direction of the quaternion, scaled by scale. 
  * 
  *****************************************************************************/
  void toDirector(int i, DclVector& d, DclFloat scale = 1.0);

  /*****************************************************************************/
  /** Rotates the vector v around q, and returns v' = q * v * q^-1
  *
  *****************************************************************************/
  DclVector rotate(DclVector& v);

  inline void inverseMatrix(DclMatrix& q_inverse);

};

// include the inline functions
#include "vector.h"
#include "quaternion.inl"

#endif

//****************************************************************************//
